@@ -1,12 +1,12 @@
 from pyfluids import Fluid, FluidsList, Input, HumidAir, InputHumidAir
 try:
-    from .variable import variable
+    from .variable import variable, scalarVariable, arrayVariable
     from .unit import unit
 except ImportError:
-    from variable import variable
+    from variable import variable, scalarVariable, arrayVariable
     from unit import unit
     
-dx = 0.00001
+dx = 0.000001
 
 def prop(property, fluid, **kwargs):
 
@@ -46,8 +46,8 @@ def differentialsBrine(fluid : Fluid, fluidName,  property, C, parameters):
         return  vars, grads
 
 
-    y1 = getattr(Fluid(fluidName,C.value*(1 + dx)).with_state(*fluid._inputs), knownProperties[property])
-    y2 = getattr(Fluid(fluidName,C.value*(1 - dx)).with_state(*fluid._inputs), knownProperties[property])
+    y1 = getattr(Fluid(fluidName,C.value*(1 + dx)).with_state(*fluid._inputs), property)
+    y2 = getattr(Fluid(fluidName,C.value*(1 - dx)).with_state(*fluid._inputs), property)
 
     grads.append((y2-y1) / (2*dx * C.value))
     vars.append(C)
@@ -75,11 +75,11 @@ def differentials(fluid : Fluid, property : str, parameters):
         
         inputs[i] = i1
         fluid.update(*inputs)
-        y1 = getattr(fluid, knownProperties[property])
+        y1 = getattr(fluid, property)
         
         inputs[i] = i2
         fluid.update(*inputs)
-        y2 = getattr(fluid, knownProperties[property])
+        y2 = getattr(fluid, property)
         
         inputs[i] = i0
         fluid.update(*inputs)
@@ -92,27 +92,28 @@ def differentials(fluid : Fluid, property : str, parameters):
 def outputFromParameters(scalarMethod, property, params):
     
     
-    if (all([elem.len() == 1 for elem in params if not elem is None])):
+    if (all([type(elem) == scalarVariable for elem in params if not elem is None])):
         ## all inputs are scalars
         out = scalarMethod(property, *params)
         return out
     
     out = []
+    ns = []
     paramVecs = [None] * len(params)
-    n = None
 
     for i, param in enumerate(params):
-        if not param is None:
-            if param.len() != 1:
-                paramVecs[i] = [variable(val, param.unit, unc) for val,unc in zip(param.value, param.uncert)]
-
-    for i in range(len(paramVecs)):
-        if not paramVecs[i] is None:
-            n = len(paramVecs[i])
+        if type(param) == arrayVariable:
+            paramVecs[i] = param
+            ns.append(len(param))
     
-    for i in range(len(paramVecs)):
-        if paramVecs[i] is None:
-            paramVecs[i] = [params[i]] * n
+    if not (all([ns[0] == n] for n in ns)):
+        raise ValueError('All parameters has to have the same length')
+    
+    
+    n = ns[0]
+    for i, param in enumerate(params):
+        if isinstance(param, scalarVariable):
+            paramVecs[i] = variable([param.value] * n, param.unit, [param.uncert] * n)
     
     for i in range(n):
         params = [elem[i] for elem in paramVecs]
@@ -149,7 +150,7 @@ def propWaterScalar(property, T,P):
     fluid = fluid.with_state(Input.temperature(T.value), Input.pressure(P.value))
     
     ## create a variable from the fluid
-    var = getattr(fluid, knownProperties[property])
+    var = getattr(fluid, property)
     var = variable(var, propertyUnits[property])
     vars, grads = differentials(fluid, property, [T, P])
     var._addDependents(vars, grads)
@@ -199,7 +200,7 @@ def propMEGScalar(property, T,P,C):
     fluid = fluid.with_state(Input.temperature(T.value), Input.pressure(P.value))
     
     ## create a variable from the fluid
-    var = getattr(fluid, knownProperties[property])
+    var = getattr(fluid, property)
     var = variable(var, propertyUnits[property])
     vars, grads = differentialsBrine(fluid, FluidsList.MEG, property, C, [T, P])
     var._addDependents(vars, grads)
@@ -273,7 +274,7 @@ def propHumidAirScalar(property, T,Rh,H = None, P = None):
     fluid = fluid.with_state(*inputs)
     
     ## create a variable from the fluid
-    var = getattr(fluid, knownProperties[property])
+    var = getattr(fluid, property)
     var = variable(var, propertyUnits[property])
     Vars, grads = differentials(fluid, property, vars)
     var._addDependents(Vars, grads)
@@ -282,16 +283,16 @@ def propHumidAirScalar(property, T,Rh,H = None, P = None):
     return var 
 
 propertyUnits = {
-    'rho': 'kg/m3',
-    'cp': 'J/kg-K',
-    'mu': 'Pa-s'
+    'density': 'kg/m3',
+    'specific_heat': 'J/kg-K',
+    'dynamic_viscosity': 'Pa-s'
 }
 
-knownProperties = {
-    'rho' : "density",
-    'cp' : "specific_heat",
-    'mu': "dynamic_viscosity"
-}
+knownProperties = [
+    "density",
+    "specific_heat",
+    "dynamic_viscosity"
+]
 
 knownFluids = {
     'water': [propWater],
@@ -301,6 +302,6 @@ knownFluids = {
 
 
 if __name__ == '__main__':
-    rho = prop('rho', 'air', P = variable(1,'bar'), T = variable([30,40,50,60,70,80],'C', [3,4,5,6,7,8]), Rh = variable(50,'%'))
-    
+    rho = prop('density', 'water', P = variable(1,'bar'), T = variable(30,'C'))
+    print(rho)
     
\ No newline at end of file
