@@ -1,9 +1,10 @@
 from scipy.optimize import minimize
 import numpy as np
+import warnings
 try:
-    from .variable import variable
+    from .variable import scalarVariable
 except ImportError:
-    from variable import variable
+    from variable import scalarVariable
 
 
 
@@ -33,7 +34,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
         if (len(o) != 2):
             raise ValueError('Each equation has to be a list of 2 elements')
         for elem in o:
-            if not isinstance(elem, variable):
+            if not isinstance(elem, scalarVariable):
                 raise ValueError('Each side of each equation has to be a variable')        
         ## test if the units match
         if (o[0]._unitObject._SIBaseUnit != o[1]._unitObject._SIBaseUnit):
@@ -66,7 +67,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
                 if len(o) != 3:
                     raise ValueError('Each bound has to have 3 elements')
                 for elem in o:
-                    if not isinstance(elem, variable):
+                    if not isinstance(elem, scalarVariable):
                         raise ValueError('Each side of the bounds has to be a variable')         
                 ## check the bounds
                 if (o[0]._unitObject._SIBaseUnit != o[1]._unitObject._SIBaseUnit or o[1]._unitObject._SIBaseUnit != o[2]._unitObject._SIBaseUnit):
@@ -83,7 +84,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
             bbounds = []
             for i, elem in enumerate(bounds):
                 for e in elem:
-                    if not isinstance(e, variable):
+                    if not isinstance(e, scalarVariable):
                         raise ValueError('All bounds has to be variables')
                     e.convert(x[i].unit)
                 bbounds.append([e.value for e in elem])
@@ -106,7 +107,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
                     for elem in bound:
                         elem.convert(elem._unitObject._SIBaseUnit)
                 var = np.min([np.max(bound[0], bound[1]), bound[2]])
-                x[boundIndexes[i]]._value = np.array([var.value], dtype = float)
+                x[boundIndexes[i]]._value = var.value
                     
     
     ## define the minimization problem
@@ -114,7 +115,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
        
         ## update the values of the variables
         for xi, xxi in zip(x,xx):
-            xi._value = np.array([xxi],dtype=float)
+            xi._value = xxi
       
         ## evaluate the function
         out = ffunc(*x)
@@ -127,9 +128,9 @@ def solve(func, x, *args, bounds = None,**kwargs):
 
         return sum([(e[0].value - e[1].value)**2 * s for e,s in zip(out, scales)])
 
-    
     ## run the minimization
     if callable(bounds): keepVariablesFeasible()
+    warnings.filterwarnings('ignore')
     out = minimize(
         minimizationProblem,
         [elem.value for elem in x],
@@ -138,6 +139,7 @@ def solve(func, x, *args, bounds = None,**kwargs):
         bounds = bounds if not callable(bounds) else None,
         callback=keepVariablesFeasible if callable(bounds) else None
         )
+    warnings.filterwarnings('default')
     if callable(bounds): keepVariablesFeasible()
 
     ## loop over all equations and create a list of the residuals and create the jacobian matrix
@@ -145,7 +147,9 @@ def solve(func, x, *args, bounds = None,**kwargs):
     for i, equation in enumerate(ffunc(*x)):
 
         ## add the residual
-        residuals.append(equation[0] - equation[1])
+        res = equation[0] - equation[1]
+        res.convert(res._unitObject._SIBaseUnit)
+        residuals.append(res)
 
         ## loop over the variables
         for j, xj in enumerate(x):
@@ -165,4 +169,24 @@ def solve(func, x, *args, bounds = None,**kwargs):
 
 
 
+if __name__=="__main__":
+    from variable import variable
+    from prop import prop
+    
+    p = variable(1,'bar', 0.01)
+    c = variable(50,'%')
+    t_in = variable(60,'C', 1.2)
+    phi = variable(120, 'kW', 2.3)
+    flow = variable(350, 'L/min', 9.1)
+    
+    def func(t):
+        t_avg = t_in + (t - t_in) / 2
+        rho = prop('density', 'MEG', P = p, T = t_avg, C = c)
+        cp = prop('specific_heat', 'MEG', P = p, T = t_avg, C = c)
+        phi_calc = rho * cp * flow * (t_in - t)
+        phi_calc.convert('kW')
+        return [phi, phi_calc]
+
+    x0 = variable(20, 'C')
+    t_out = solve(func, x0)
 
