@@ -1,16 +1,22 @@
 import numpy as np
 try:
-    from .unitSystem import knownCharacters, knownPrefixes, knownUnits, baseUnit, _unitConversion, knownUnitsDict
+    from .unitSystem import _unitConversion
+    from .__init__ import knownPrefixes, knownUnits, knownCharacters, knownUnitsDict, baseUnit
 except ImportError:
-    from unitSystem import knownCharacters, knownPrefixes, knownUnits, baseUnit, _unitConversion, knownUnitsDict
-    
+    from unitSystem import _unitConversion, getDicts
+    knownUnits, knownCharacters, knownUnitsDict, knownPrefixes, baseUnit = getDicts()
+
+hyphen = '-'
+slash = '/'
+integers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
+temperatureUnits = knownUnitsDict['K']
 
 class unit():
 
     def __init__(self, unitStr) -> None:
         if unitStr == '':
             unitStr = '1'
-
+        
         # remove any unknown characters
         unitStr = self._formatUnit(unitStr)
 
@@ -21,7 +27,8 @@ class unit():
         self.unitStr = self._createUnitString()
 
         self._SIBaseUnit = self._getSIBaseUnit(self.upper, self.upperExp, self.lower, self.lowerExp)
-        self._converterToSI = self.getConverter(self._SIBaseUnit)
+        otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp = self._getLists(self._SIBaseUnit)
+        self._converterToSI = self._getConverter(otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp)
 
     def _createUnitString(self):
         return self._combineUpperAndLower(self.upper, self.upperPrefix, self.upperExp, self.lower, self.lowerPrefix, self.lowerExp)
@@ -84,9 +91,9 @@ class unit():
         lower = [pre + low + exp for pre, low, exp in zip(lowerPrefix, lower, lowerExp) if low != "1"]
 
         # create a unit string
-        u = '-'.join(upper) if upper else "1"
+        u = hyphen.join(upper) if upper else "1"
         if lower:
-            lower = '-'.join(lower)
+            lower = hyphen.join(lower)
             u = u + '/' + lower
 
         return u
@@ -145,25 +152,30 @@ class unit():
 
         return out
 
+    @staticmethod
+    def _splitUnitExponentAndPrefix(listOfUnitStrings):
+        n = len(listOfUnitStrings)
+        uList = [None] * n
+        prefixList = [None] * n
+        expList = [None] * n
+        for i in range(n):
+            prefixUnit, exponent = unit._removeExponentFromUnit(listOfUnitStrings[i])
+            u, prefix = unit._removePrefixFromUnit(prefixUnit)
+            uList[i] = u
+            prefixList[i] = prefix
+            expList[i] = exponent
+        return uList, prefixList, expList
+
     @staticmethod
     def _getLists(unitStr):
         upper, lower = unit._splitCompositeUnit(unitStr)
 
-        def splitUnitExponentAndPrefix(unitList):
-            tmp = [unit._removeExponentFromUnit(elem) for elem in unitList]
-            exponent = [elem[1] for elem in tmp]
-            tmp = [elem[0]for elem in tmp]
-            tmp = [unit._removePrefixFromUnit(elem) for elem in tmp]
-            u = [elem[0] for elem in tmp]
-            prefix = [elem[1] for elem in tmp]
-            return u, prefix, exponent
-        upper, upperPrefix, upperExp = splitUnitExponentAndPrefix(upper)
-        lower, lowerPrefix, lowerExp = splitUnitExponentAndPrefix(lower)
+        upper, upperPrefix, upperExp = unit._splitUnitExponentAndPrefix(upper)
+        lower, lowerPrefix, lowerExp = unit._splitUnitExponentAndPrefix(lower)
 
-        if len(upper + lower) > 1:
-            temp = knownUnitsDict['K']
-            upper = ['DELTA' + elem if elem in temp else elem for elem in upper]
-            lower = ['DELTA' + elem if elem in temp else elem for elem in lower]
+        if lower:
+            upper = ['DELTA' + elem if elem in temperatureUnits else elem for elem in upper]
+            lower = ['DELTA' + elem if elem in temperatureUnits else elem for elem in lower]
 
         return upper, upperPrefix, upperExp, lower, lowerPrefix, lowerExp
 
@@ -177,26 +189,139 @@ class unit():
             if s not in knownCharacters:
                 raise ValueError(f'The character {s} is not used within the unitsystem')
 
-        return unitStr
+        ## find start parenthesis is the unit string
+        startParenIndexes = [i for i, s in enumerate(unitStr) if s == '(']
+        stopParenIndexes = [i for i, s in enumerate(unitStr) if s == ')']
 
+        ## return if no parenthesis where found
+        if not startParenIndexes and not stopParenIndexes:
+            return unitStr
+      
+        ## check that the number of start and stop parenthesis are equal
+        if len(startParenIndexes) != len(stopParenIndexes):
+            raise ValueError('The unit string has to have an equal number of open parenthesis and clsoe parenthesis')
+        
+        ## find all parenthesis pairs
+        allIndexes = startParenIndexes + stopParenIndexes
+        allIndexes.sort()
+        isStartParen = [elem in startParenIndexes for elem in allIndexes]
+        done, iter, maxIter, order  = False, 0, len(allIndexes), []
+        while not done:
+            for i in range(len(allIndexes)-1):
+                if isStartParen[i] and (isStartParen[i] + isStartParen[i+1] == 1):
+                    order.append([allIndexes[i], allIndexes[i+1]])
+                    allIndexes.pop(i+1)
+                    allIndexes.pop(i)
+                    isStartParen.pop(i+1)
+                    isStartParen.pop(i)
+                    break
+
+            if len(allIndexes) == 0: break
+            
+            iter += 1
+            if iter > maxIter:
+                raise ValueError('An order to evaluate the parenthesis could not be found')
+        
+
+        ## find the outer most parenthesis        
+        done, maxIter, iter = False, len(order), 0
+        while not done:
+            changesMade = False
+            for i in range(len(order)):
+                iLow = order[i][0]
+                iHigh = order[i][1]
+                for j in range(len(order)):
+                    jLow = order[j][0]
+                    jHigh = order[j][1]
+                    if iLow < jLow and iHigh > jHigh:
+                        order.pop(j)
+                        changesMade = True
+                        break
+                if changesMade:
+                    break
+            if not changesMade:
+                break
+            iter += 1
+            if iter > maxIter:
+                raise ValueError('An order to evaluate the parenthesis could not be found')
+        
+        ## find all slashes outside the outer most parenthesis
+        slashes = [i for i, s in enumerate(unitStr) if s == slash]
+
+        iter, maxIter, done = 0, len(slashes), False
+        while not done:
+            done = True
+            for s in slashes:    
+                for part in order:
+                    if part[0] < s < part[1]:
+                        done = False
+                        slashes.remove(s)
+                        break                
+            if done: break
+            iter += 1
+            if iter > maxIter:
+                raise ValueError('An order to evaluate the parenthesis could not be found')
+
+        if len(slashes) > 1:
+            raise ValueError('You have more than one slash ("/") outside the outer most parenthesis. This makes the unit ambiguis')
+        
+        if slashes:
+            splits = [-1] + slashes
+            parts = [unitStr[i+1:j] for i,j in zip(splits, splits[1:]+[None])]
+            parts = [unit._formatUnit(part) for part in parts]
+            if isinstance(parts, list):
+                return '/'.join(parts)
+            return parts
+        
+        parts = unit._formatUnit(unitStr[order[0][0]+1: order[0][1]])
+        if order[0][1] != len(unitStr) - 1:
+            remaining = unitStr[order[0][1]+1:]
+            parts = parts.split('/')
+            if len(parts) > 1:
+                upper, lower = parts
+                lower = lower.split('-')
+            else:
+                upper, lower = parts[0], []
+            upper = upper.split('-')
+            try: 
+                exponent = int(remaining)
+                for i, up in enumerate(upper):
+                    u, e = unit._removeExponentFromUnit(up)
+                    upper[i] = u + str(e * exponent)
+                for i, low in enumerate(lower):
+                    u, e = unit._removeExponentFromUnit(low)
+                    lower[i] = u + str(e * exponent)
+                out = '-'.join(upper)
+                if lower:
+                    out += '/' + '-'.join(lower)
+                return out
+            except TypeError:
+                raise TypeError()
+        return parts
+    
     @ staticmethod
     def _splitCompositeUnit(compositeUnit):
-        compositeUnit = compositeUnit.split('/')
+        lower = []
+        if not slash in compositeUnit:
+            upper = compositeUnit.split(hyphen)
+            return upper, lower
+    
+        compositeUnit = compositeUnit.split(slash)
 
         if len(compositeUnit) > 2:
             raise ValueError('A unit can only have a single slash (/)')
 
-        upper = compositeUnit[0].split('-')
-        lower = compositeUnit[1].split('-') if len(compositeUnit) > 1 else []
+        upper = compositeUnit[0].split(hyphen)
+        lower = compositeUnit[1].split(hyphen) if len(compositeUnit) > 1 else []
 
         return upper, lower
 
     @ staticmethod
     def _removeExponentFromUnit(u):
-        u = list(u)
-        lenU = len(u)
+        integerIndexes = [i for i, char in enumerate(u) if char in integers]
         exponent = 1
-        integerIndexes = [i for i, char in enumerate(u) if char in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']]
+        if not integerIndexes:
+            return u, exponent
 
         # override the exponent if there are any integerindexes
         if integerIndexes:
@@ -206,15 +331,13 @@ class unit():
             if sum(integerIndexes) != (maxIndex * (maxIndex + 1) - minIndex * (minIndex + 1)) / 2:
                 raise ValueError('All numbers in the unit has to be grouped together')
 
-            # Determien if the last integer is placed at the end of the unit
-            if integerIndexes[-1] != lenU - 1:
+            # Determine if the last integer is placed at the end of the unit
+            if integerIndexes[-1] != len(u) - 1:
                 raise ValueError('Any number has to be placed at the end of the unit')
 
-            # join the integers
-            exponent = int(''.join([u[i] for i in integerIndexes]))
+            exponent = int(u[minIndex+1:])
+            u = u[:minIndex+1]
 
-        # join the unit
-        u = ''.join([u[i] for i in range(lenU) if i not in integerIndexes])
 
         # Ensure that the entire use was not removed by removing the integers
         if not u:
@@ -222,51 +345,28 @@ class unit():
             if exponent != 1:
                 raise ValueError(f'The unit {u} was stripped of all integers which left no symbols in the unit. This is normally due to the integers removed being equal to 1, as the unit is THE unit. Howver, the intergers removed was not equal to 1. The unit is therefore not known.')
             u = '1'
-
         return u, exponent
 
     @staticmethod
     def _assertEqualStatic(a, b):
 
-        aUpper, aUpperPrefix, aUpperExp, aLower, aLowerPrefix, aLowerExp = unit._getLists(a)
-        bUpper, bUpperPrefix, bUpperExp, bLower, bLowerPrefix, bLowerExp = unit._getLists(b)
-
-        if bool(aLower) != bool(bLower):
+        aUpper, aLower = unit._splitCompositeUnit(a)
+        bUpper, bLower = unit._splitCompositeUnit(b)
+               
+        for elem in aUpper:
+            if not elem in bUpper:
+                return False
+            bUpper.remove(elem)
+        if bUpper:
             return False
-
-        aUpperIndexes = np.argsort(aUpper)
-        aLowerIndexes = np.argsort(aLower)
-        bUpperIndexes = np.argsort(bUpper)
-        bLowerIndexes = np.argsort(bLower)
-
-        aUpperSorted = list(np.sort(aUpper))
-        aLowerSorted = list(np.sort(aLower))
-        bUpperSorted = list(np.sort(bUpper))
-        bLowerSorted = list(np.sort(bLower))
-
-        aUpperExpSorted = [aUpperExp[elem] for elem in aUpperIndexes]
-        aLowerExpSorted = [aLowerExp[elem] for elem in aLowerIndexes]
-        bUpperExpSorted = [bUpperExp[elem] for elem in bUpperIndexes]
-        bLowerExpSorted = [bLowerExp[elem] for elem in bLowerIndexes]
-
-        aUpperPrefixSorted = [aUpperPrefix[elem] for elem in aUpperIndexes]
-        aLowerPrefixSorted = [aLowerPrefix[elem] for elem in aLowerIndexes]
-        bUpperPrefixSorted = [bUpperPrefix[elem] for elem in bUpperIndexes]
-        bLowerPrefixSorted = [bLowerPrefix[elem] for elem in bLowerIndexes]
-
-        if aUpperSorted != bUpperSorted:
+        
+        for elem in aLower:
+            if not elem in bLower:
+                return False
+            bLower.remove(elem)
+        if bLower:
             return False
-        if aLowerSorted != bLowerSorted:
-            return False
-        if aUpperExpSorted != bUpperExpSorted:
-            return False
-        if aLowerExpSorted != bLowerExpSorted:
-            return False
-        if aUpperPrefixSorted != bUpperPrefixSorted:
-            return False
-        if aLowerPrefixSorted != bLowerPrefixSorted:
-            return False
-
+        
         return True
 
     @staticmethod
@@ -303,60 +403,38 @@ class unit():
         return unit, prefix
 
     @staticmethod
-    def _getSIBaseUnit(upper, upperExp, lower, lowerExp):
-
-        def addUnitsMultipleTimes(units, unitExps):
-            out = []
-            for unit, unitExp in zip(units, unitExps):
-                for _ in range(unitExp):
-                    out.append(unit)
-            return out
-
-        upper = addUnitsMultipleTimes(upper, upperExp)
-        lower = addUnitsMultipleTimes(lower, lowerExp)
-
-        # remove the prefix - this is just a scaling so this does not change the SI base unit
-        upper = [unit._removePrefixFromUnit(elem)[0] for elem in upper]
-        lower = [unit._removePrefixFromUnit(elem)[0] for elem in lower]
-
-        # return the splitted version of the base unit
-        upper = [unit._splitCompositeUnit(knownUnits[elem][0]) for elem in upper]
-        lower = [unit._splitCompositeUnit(knownUnits[elem][0]) for elem in lower]
-
-        # combine the upper and lower
-        tmpUpper = [elem[0] for elem in upper] + [elem[1] for elem in lower]
-        tmpLower = [elem[1] for elem in upper] + [elem[0] for elem in lower]
-        upper = [elem for L in tmpUpper for elem in L]
-        lower = [elem for L in tmpLower for elem in L]
-
-        # remove the exponents
-        tmpUpper = [unit._removeExponentFromUnit(elem) for elem in upper]
-        tmpLower = [unit._removeExponentFromUnit(elem) for elem in lower]
-        upper = [elem[0] for elem in tmpUpper]
-        tmpUpperExp = [elem[1]for elem in tmpUpper]
-        lower = [elem[0] for elem in tmpLower]
-        tmpLowerExp = [elem[1]for elem in tmpLower]
-
-        upperSet = list(set(upper))
-        upperSet.sort()
-        lowerSet = list(set(lower))
-        lowerSet.sort()
-        lenUpperSet = len(upperSet)
-        lenLowerSet = len(lowerSet)
-        upperExp = [0] * lenUpperSet
-        lowerExp = [0] * lenLowerSet
-        upperPrefix = [None] * lenUpperSet
-        lowerPrefix = [None] * lenLowerSet
-        for i, u in enumerate(upperSet):
-            indexes = [_ for _, elem in enumerate(upper) if elem == u]
-            upperExp[i] = sum([tmpUpperExp[elem] for elem in indexes])
-        for i, l in enumerate(lowerSet):
-            indexes = [_ for _, elem in enumerate(lower) if elem == l]
-            lowerExp[i] = sum([tmpLowerExp[elem] for elem in indexes])
-        upper, lower = upperSet, lowerSet
+    def _getSIBaseUnit(upper, upperExp, lower, lowerExp):       
+        nUpper = len(upper)
+        upperOut = []
+        lowerOut = []
+        upperExpOut = []
+        lowerExpOut = []
+        for i, (elem, exp) in enumerate(zip(upper + lower, upperExp + lowerExp)):
+            up,low = unit._splitCompositeUnit(knownUnits[elem][0])
+            if (i >= nUpper):
+                up, low = low, up
+            for elem in up:
+                elem, elemExp = unit._removeExponentFromUnit(elem)
+                if not elem in upperOut:
+                    upperOut.append(elem)
+                    upperExpOut.append(elemExp * exp)
+                else:
+                    index = upperOut.index(elem)
+                    upperExpOut[index] += elemExp * exp
+            for elem in low:
+                elem, elemExp = unit._removeExponentFromUnit(elem)
+                if not elem in lowerOut:
+                    lowerOut.append(elem)
+                    lowerExpOut.append(elemExp * exp)
+                else:
+                    index = lowerOut.index(elem)
+                    lowerExpOut[index] += elemExp * exp
 
+        upperPrefix = [None] * len(upperOut)
+        lowerPrefix = [None] * len(lowerOut)
+        
         upper, upperPrefix, upperExp, lower, lowerPrefix, lowerExp = unit._cancleUnits(
-            upper, upperPrefix, upperExp, lower, lowerPrefix, lowerExp
+            upperOut, upperPrefix, upperExpOut, lowerOut, lowerPrefix, lowerExpOut
         )
 
         return unit._combineUpperAndLower(upper, upperPrefix, upperExp, lower, lowerPrefix, lowerExp)
@@ -432,7 +510,7 @@ class unit():
         # test if the units are identical
         if self._assertEqual(other):
             if self._SIBaseUnit == 'K':
-                return True, 'DELTA' + self.unitStr
+                return True, 'DELTAK'
             return True, self._SIBaseUnit
 
         # test if the SI base units are identical
@@ -451,13 +529,13 @@ class unit():
         return unit._multiply(self.unitStr, other.unitStr)
 
     def __truediv__(self, other):
-
-        if self._SIBaseUnit == 'K':
+        
+        if self._SIBaseUnit == 'K' and other._SIBaseUnit != '1':
             a = 'DELTA' + self.unitStr
         else:
             a = self.unitStr
 
-        if other._SIBaseUnit == 'K':
+        if other._SIBaseUnit == 'K' and self._SIBaseUnit != '1':
             b = 'DELTA' + other.unitStr
         else:
             b = other.unitStr
@@ -533,7 +611,20 @@ class unit():
 
         if unit._assertEqualStatic(self._SIBaseUnit, otherSIBase) == False:
             raise ValueError(f'You tried to convert from {self} to {newUnit}. But these do not have the same base units')
+        
+        return self._getConverter(otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp)
 
+    def getUnitWithoutPrefix(self):
+        return self._removePrefixFromUnit(self.unitStr)[0]
+
+    def getLogarithmicConverter(self, unitStr = None):
+        if unitStr is None:
+            unitStr = self.unitStr
+        u, p = self._removePrefixFromUnit(unitStr)
+        return knownUnits[u][1]
+   
+    def _getConverter(self, otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp):
+        
         # initialize the scale and offset
         out = _unitConversion(1, 0)
 
@@ -543,17 +634,17 @@ class unit():
 
         # modify the scale and offset using the conversions
         conversions = upperConversions + lowerConversions
-        conversionBool = [True] * len(upperConversions) + [False] * len(lowerConversions)
+        nUpper = len(upperConversions)
         prefixes = self.upperPrefix + self.lowerPrefix
         exponents = self.upperExp + self.lowerExp
-        for conv, prefix, exp, upperBool in zip(conversions, prefixes, exponents, conversionBool):
+        for i, (conv, prefix, exp) in enumerate(zip(conversions, prefixes, exponents)):
+            if exp > 1: conv = conv ** exp
             if not prefix is None:
                 conv *= knownPrefixes[prefix]
-            for _ in range(exp):
-                if upperBool:
-                    out *= conv
-                else:
-                    out /= conv
+            if i < nUpper:
+                out *= conv
+            else:
+                out /= conv
 
         # get all conversions from the upper and lower units in the new unit
         upperConversions = [knownUnits[elem][1] for elem in otherUpper]
@@ -561,19 +652,17 @@ class unit():
 
         # modify the scale and offset based on the conversions
         conversions = upperConversions + lowerConversions
-        conversionBool = [True] * len(upperConversions) + [False] * len(lowerConversions)
+        nUpper = len(otherUpper)
         prefixes = otherUpperPrefix + otherLowerPrefix
         exponents = otherUpperExp + otherLowerExp
-        for conv, prefix, exp, upperBool in zip(conversions, prefixes, exponents, conversionBool):
+        for i, (conv, prefix, exp) in enumerate(zip(conversions, prefixes, exponents)):
+            if exp > 1: conv = conv ** exp
             if not prefix is None:
                 conv *= knownPrefixes[prefix]
-
-            for _ in range(exp):
-                # the multiply and divisions are swapped because the conversion is away from the SI unit system
-                if upperBool:
-                    out /= conv
-                else:
-                    out *= conv
+            if i < nUpper:
+                out /= conv
+            else:
+                out *= conv
 
         return out
 
@@ -591,7 +680,12 @@ class unit():
         self.unitStr = self._createUnitString()
 
         self._SIBaseUnit = self._getSIBaseUnit(self.upper, self.upperExp, self.lower, self.lowerExp)
-        self._converterToSI = self.getConverter(self._SIBaseUnit)
+        otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp = self._getLists(self._SIBaseUnit)
+        self._converterToSI = self._getConverter(otherUpper, otherUpperPrefix, otherUpperExp, otherLower, otherLowerPrefix, otherLowerExp)
 
+if __name__ == "__main__":
+    a = unit._formatUnit('(m-s2)2 / Hz')
+    print(a)
+    
 
 
