@@ -5,11 +5,176 @@ try:
 except ImportError:
     from unit import unit
     
-HANDLED_FUNCTIONS = {}
 
+class temperatureVariable:
+    def __init__(self):
+        pass
+    
+    @staticmethod
+    def addTemperatures(a, b):
 
-class variable():
-    def __init__(self, value, unitStr='', uncert=None, nDigits=3) -> None:
+        # convert self and other to the SI unit system
+        selfUnit = deepcopy(a.unit)
+        otherUnit = deepcopy(b.unit)
+        a.convert(a._unitObject._SIBaseUnit)
+        b.convert(b._unitObject._SIBaseUnit)
+
+        # determine the value and gradients
+        val = a._value + b._value
+        grad = [1, 1]
+        vars = [a, b]
+
+        # create the new variable
+        var = variable(val, 'K')
+        var._addDependents(vars, grad)
+        var._calculateUncertanty()
+
+        # convert self and other back
+        a.convert(selfUnit)
+        b.convert(otherUnit)
+        
+
+        SIBaseUnits = [a._unitObject._SIBaseUnit, b._unitObject._SIBaseUnit]
+        if 'DELTAK' in SIBaseUnits:
+            outputUnit = [a.unit, b.unit][SIBaseUnits.index('K')]
+            var.convert(outputUnit)
+        else:
+            if selfUnit == otherUnit:
+                var.convert(selfUnit)
+        return var
+    
+    def subtractTemperatures(a,b):
+
+        # convert self and other to the SI unit system
+        aUnit = deepcopy(a.unit)
+        bUnit = deepcopy(b.unit)
+        
+        SIBaseUnits = [a._unitObject._SIBaseUnit, b._unitObject._SIBaseUnit]
+        if (SIBaseUnits == ['DELTAK', 'K']):
+            raise ValueError('You can not subtract a temperature from a tempeature difference')
+        
+        nTemp = sum([1 for elem in SIBaseUnits if elem == 'K'])
+        outputUnit = 'K' if nTemp == 1 else 'DELTAK'
+        
+        a.convert(a._unitObject._SIBaseUnit)
+        b.convert(b._unitObject._SIBaseUnit)
+
+        # determine the value and gradients
+        val = a.value - b.value
+        grad = [1, -1]
+        vars = [a, b]
+
+        # create the new variable
+        var = variable(val, outputUnit)
+        var._addDependents(vars, grad)
+        var._calculateUncertanty()
+
+        # convert self and other back
+        a.convert(aUnit)
+        b.convert(bUnit)
+
+        # convert the output variable
+        if 'DELTAK' in SIBaseUnits and 'K' in SIBaseUnits:
+            outputUnit = [a.unit, b.unit][SIBaseUnits.index('K')]
+            var.convert(outputUnit)
+        elif aUnit == bUnit:
+            if SIBaseUnits[0] == 'DELTAK':
+                var.convert(aUnit)
+            else:
+                var.convert('DELTA' + aUnit)
+        
+        return var
+    
+class logarithmicVariables:
+    def __init__(self):
+        pass
+    
+    @staticmethod
+    def __add__(a,b):
+        aUnit = str(a.unit)
+        bUnit =str(b.unit)
+        
+        aUnitWithoutPrefix = a._unitObject.getUnitWithoutPrefix()
+        a.convert(aUnitWithoutPrefix)
+
+        aConverter = a._unitObject.getLogarithmicConverter()
+        a._unitStr = '1'
+        a._unitObject = unit('')
+        aConverter.convertToSignal(a)
+       
+        bUnitWithoutPrefix = b._unitObject.getUnitWithoutPrefix()
+        b.convert(bUnitWithoutPrefix)
+        bConverter = b._unitObject.getLogarithmicConverter()
+        b._unitStr = '1'
+        b._unitObject = unit('')
+        bConverter.convertToSignal(b)
+
+        c = a + b
+     
+        aConverter.convertFromSignal(a)
+        a._unitStr = aUnitWithoutPrefix
+        a._unitObject = unit(aUnitWithoutPrefix)
+        a.convert(aUnit)
+        
+        bConverter.convertFromSignal(b)
+        b._unitStr = bUnitWithoutPrefix
+        b._unitObject = unit(bUnitWithoutPrefix)
+        b.convert(bUnit)
+        
+        
+        aConverter.convertFromSignal(c)
+        c._unitStr = aUnitWithoutPrefix
+        c._unitObject = unit(aUnitWithoutPrefix)
+        c.convert(aUnit)
+
+        return c
+
+    @staticmethod
+    def __sub__(a,b):
+        aUnit = str(a.unit)
+        bUnit =str(b.unit)
+        
+        aUnitWithoutPrefix = a._unitObject.getUnitWithoutPrefix()
+        a.convert(aUnitWithoutPrefix)
+
+        aConverter = a._unitObject.getLogarithmicConverter()
+        a._unitStr = '1'
+        a._unitObject = unit('')
+        aConverter.convertToSignal(a)
+       
+        bUnitWithoutPrefix = b._unitObject.getUnitWithoutPrefix()
+        b.convert(bUnitWithoutPrefix)
+        bConverter = b._unitObject.getLogarithmicConverter()
+        b._unitStr = '1'
+        b._unitObject = unit('')
+        bConverter.convertToSignal(b)
+
+        c = a - b
+     
+        aConverter.convertFromSignal(a)
+        a._unitStr = aUnitWithoutPrefix
+        a._unitObject = unit(aUnitWithoutPrefix)
+        a.convert(aUnit)
+        
+        bConverter.convertFromSignal(b)
+        b._unitStr = bUnitWithoutPrefix
+        b._unitObject = unit(bUnitWithoutPrefix)
+        b.convert(bUnit)
+        
+        
+        aConverter.convertFromSignal(c)
+        c._unitStr = aUnitWithoutPrefix
+        c._unitObject = unit(aUnitWithoutPrefix)
+        c.convert(aUnit)
+
+        return c
+   
+
+class scalarVariable():
+    def __init__(self, value, unitStr, uncert, nDigits) -> None:
+        
+        self._value = value
+        self._uncert = uncert
 
         # create a unit object
         self._unitObject = unitStr if isinstance(unitStr, unit) else unit(unitStr)
@@ -17,31 +182,6 @@ class variable():
         # number of digits to show
         self.nDigits = nDigits
 
-        # store the value and the uncertaty
-        def evaluateInput(input):
-            if isinstance(input, np.ndarray):
-                return input
-            else:
-                listLike = False
-                try:
-                    len(input)
-                    listLike = True
-                except:
-                    pass
-                if listLike:
-                    return np.array(input, dtype=float)
-                else:
-                    return np.array([input], dtype=float)
-        self._value = evaluateInput(value)
-        if uncert is None:
-            self._uncert = evaluateInput([0] * self.len())
-        else:
-            self._uncert = evaluateInput(uncert)
-
-        # check the length of the uncertanty and the value
-        if self.len() != len(self._uncert):
-            raise ValueError('The lenght of the value has to be equal to the lenght of the uncertanty')
-
         # value and unit in SI. This is used when determining the gradient in the uncertanty expression
         self._getConverterToSI()
 
@@ -54,8 +194,6 @@ class variable():
 
     @property
     def value(self):
-        if self.len() == 1:
-            return self._value[0]
         return self._value
 
     @property
@@ -64,8 +202,6 @@ class variable():
 
     @property
     def uncert(self):
-        if self.len() == 1:
-            return self._uncert[0]
         return self._uncert
 
     def convert(self, newUnit):
@@ -76,24 +212,7 @@ class variable():
 
         # update the converter to SI
         self._getConverterToSI()
-
-    def len(self):
-        return len(self._value)
-
-    def __getitem__(self, index):
-        if isinstance(index, int) or isinstance(index, slice):
-            if index >= self.len():
-                raise IndexError('Index out of bounds')
-            if self.len() == 1:
-                if index != 0:
-                    raise IndexError('Index out of bound')
-                return variable(self.value, self._unitObject, self.uncert)
-            return variable(self.value[index], self._unitObject, self.uncert[index])
-        else:
-            val = [self.value[elem]for elem in index]
-            unc = [self.uncert[elem]for elem in index]
-            return variable(val, self._unitObject, unc)
-
+ 
     def printUncertanty(self, value, uncert):
         # function to print number
         if uncert == 0 or uncert is None or np.isnan(uncert):
@@ -149,55 +268,17 @@ class variable():
         else:
             unitStr = rf'{squareBracketLeft}{unitStr}{squareBracketRight}'
 
-        if self.len() == 1:
-            # print a single value
-            value = self.value
-            if self.uncert != 0:
-                uncert = self.uncert
 
-            value, uncert = self.printUncertanty(value, uncert)
-            if uncert is None:
-                return rf'{value}{space}{unitStr}'
-            else:
-                return rf'{value} {pm} {uncert}{space}{unitStr}'
+        # print a single value
+        value = self.value
+        if self.uncert != 0:
+            uncert = self.uncert
 
+        value, uncert = self.printUncertanty(value, uncert)
+        if uncert is None:
+            return rf'{value}{space}{unitStr}'
         else:
-            # print array of values
-            valStr = []
-            uncStr = []
-            for v, u in zip(self._value, self._uncert):
-                v, u = self.printUncertanty(v, u)
-                valStr.append(v)
-                uncStr.append(u)
-
-            if all(self._uncert == 0) or all(elem is None for elem in self._uncert):
-                out = rf''
-                out += rf'['
-                for i, elem in enumerate(valStr):
-                    out += rf'{elem}'
-                    if i != len(valStr) - 1:
-                        out += rf', '
-                out += rf']'
-                out += rf'{space}{unitStr}'
-                return out
-            else:
-                # find number of significant digits in uncertanty
-                out = rf''
-                out += rf'['
-                for i, elem in enumerate(valStr):
-                    out += rf'{elem}'
-                    if i != len(valStr) - 1:
-                        out += r', '
-                out += rf']'
-                out += rf' {pm} '
-                out += rf'['
-                for i, elem in enumerate(uncStr):
-                    out += rf'{elem}'
-                    if i != len(uncStr) - 1:
-                        out += r', '
-                out += rf']'
-                out += rf'{space}{unitStr}'
-                return out
+                return rf'{value} {pm} {uncert}{space}{unitStr}'
 
     def _addDependents(self, vars, grads):
         # copy the gradients
@@ -235,15 +316,12 @@ class variable():
     def _calculateUncertanty(self):
 
         # variance from each measurement
-        if self.len() == 0:
-            variance = 0
-        else:
-            variance = np.zeros(self.len())
+        variance = 0
         selfScaleToSI = self._converterToSI.convert(1, useOffset=False)
         for var, grad in self.dependsOn.items():
             # the gradient is scaled with the inverse of the conversion of the unit to SI units.
             # This is necessary if the variable "var" has been converted after the dependency has been noted
-            scale = var._converterToSI.convert(1, useOffset=False) / selfScaleToSI
+            scale = var._converterToSI.convert(1, useOffset=False) / selfScaleToSI           
             variance += (grad * scale * var.uncert)**2
 
         # variance from the corralation between measurements
@@ -265,14 +343,22 @@ class variable():
         
     def __add__(self, other):
         
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self + variable(other, self.unit)
 
         # determine if the two variable can be added
         addBool, outputUnit = self._unitObject + other._unitObject
+        
         if not addBool:
             raise ValueError(f'You tried to add a variable in [{self.unit}] to a variable in [{other.unit}], but the units do not have the same SI base unit')
 
+        ## handle temperatures diferently
+        if outputUnit == 'K':
+            return temperatureVariable.addTemperatures(self, other)
+        
+        if outputUnit == 'Np':
+            return logarithmicVariables.__add__(self, other)
+
         # convert self and other to the SI unit system
         selfUnit = deepcopy(self.unit)
         otherUnit = deepcopy(other.unit)
@@ -289,32 +375,38 @@ class variable():
         var._addDependents(vars, grad)
         var._calculateUncertanty()
 
-        # convert self and other back
+        # convert all units back to their original units
         self.convert(selfUnit)
         other.convert(otherUnit)
-
-        # convert the output variable
-        if outputUnit == 'K':
-            SIBaseUnits = [self._unitObject._SIBaseUnit, other._unitObject._SIBaseUnit]
-            if 'DELTAK' in SIBaseUnits:
-                outputUnit = [self.unit, other.unit][SIBaseUnits.index('K')]
-                var.convert(outputUnit)
-        if self.unit == other.unit:
-            var.convert(self.unit)
+        
+        ## convert the variable in to the original unit if self and other has the same original unit
+        ## otherwise keep the variable in the SI unit system
+        if (selfUnit == otherUnit):
+            var.convert(selfUnit)
+        
         return var
 
     def __radd__(self, other):
         return self + other
 
     def __sub__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self - variable(other, self.unit)
 
         # determine if the variables can be subtracted
         subBool, outputUnit = self._unitObject - other._unitObject
+        
         if not subBool:
             raise ValueError(f'You tried to subtract a variable in [{other.unit}] from a variable in [{self.unit}], but the units do not have the same SI base unit')
 
+        ## handle temperatures diferently
+        if outputUnit == 'DELTAK' or outputUnit == 'K':
+            return temperatureVariable.subtractTemperatures(self, other)
+
+        if outputUnit == 'Np':
+            return logarithmicVariables.__sub__(self, other)
+
+
         # convert self and other to the SI unit system
         selfUnit = deepcopy(self.unit)
         otherUnit = deepcopy(other.unit)
@@ -335,13 +427,7 @@ class variable():
         self.convert(selfUnit)
         other.convert(otherUnit)
 
-        # convert the output variable
-        if outputUnit == 'K':
-            SIBaseUnits = [self._unitObject._SIBaseUnit, other._unitObject._SIBaseUnit]
-            if 'DELTAK' in SIBaseUnits:
-                outputUnit = [self.unit, other.unit][SIBaseUnits.index('K')]
-                var.convert(outputUnit)
-        if self.unit == other.unit and 'DELTA' not in outputUnit:
+        if self.unit == other.unit:
             var.convert(self.unit)
 
         return var
@@ -350,7 +436,7 @@ class variable():
         return - self + other
 
     def __mul__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self * variable(other)
 
         val = self.value * other.value
@@ -372,28 +458,30 @@ class variable():
         return self * other
 
     def __pow__(self, other):
-        if not isinstance(other, variable):
-            return self ** variable(other)
+        if not isinstance(other, scalarVariable):
+            return self ** variable(other, '')
 
-        if other.len() != 1:
-            raise ValueError('The exponent has to be a single number')
         if str(other.unit) != '1':
             raise ValueError('The exponent can not have a unit')
 
         val = self._value ** other.value
         outputUnit = self._unitObject ** other.value
 
-        gradSelf = other._value * self._value ** (other._value - 1)
+        def gradSelf(valSelf, valOTher):
+            if valSelf != 0:
+                return  valOTher * valSelf ** (valOTher - 1)
+            return 0
+        
         def gradOther(valSelf, valOther, uncertOther):
             if uncertOther != 0:
                 return valSelf ** valOther * np.log(valSelf)
-            else:
-                return 0
+            return 0
+        
         gradOther = np.vectorize(gradOther, otypes=[float])(self._value, other._value, other._uncert)
-
+        gradSelf = np.vectorize(gradSelf, otypes=[float])(self._value, other._value)
         grad = [gradSelf, gradOther]
         vars = [self, other]
-
+                
         var = variable(val, outputUnit)
         var._addDependents(vars, grad)
         var._calculateUncertanty()
@@ -403,7 +491,7 @@ class variable():
         return variable(other, '1') ** self
 
     def __truediv__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self / variable(other)
 
         val = self._value / other._value
@@ -411,7 +499,7 @@ class variable():
 
         grad = [1 / other._value, -self._value / (other._value**2)]
         vars = [self, other]
-
+        
         var = variable(val, outputUnit)
         var._addDependents(vars, grad)
         var._calculateUncertanty()
@@ -422,7 +510,7 @@ class variable():
         return var
 
     def __rtruediv__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return variable(other) / self
 
         val = other._value / self._value
@@ -542,20 +630,23 @@ class variable():
         return variable(np.abs(self.value), self.unit, self.uncert)
 
     def __array_function__(self, func, types, args, kwargs):
-        if func not in HANDLED_FUNCTIONS:
-            return NotImplemented
-        # Note: this allows subclasses that don't override
-        # __array_function__ to handle Physical objects
-        if not all(issubclass(t, variable) for t in types):
-            return NotImplemented
-        return HANDLED_FUNCTIONS[func](*args, **kwargs)
-
-    def _compareLengths(self, other):
-        if self.len() != other.len():
-            raise ValueError(f'You cannot compare {self} and {other} as they do not have the same length')
+        match func:
+            case np.max:
+                return self.max()
+            case np.min:
+                return self.min()
+            case np.mean:
+                return self.mean()
+    
+    def max(self):
+        return self
+    def min(self):
+        return self
+    def mean(self):
+        return self
 
     def __lt__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self < variable(other, self.unit)
 
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
@@ -567,22 +658,14 @@ class variable():
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
 
-        if self.len() == 1 and other.len() == 1:
-            out =  self.value < other.value
-        elif self.len() == 1:
-            out = list(self.value < other.value)
-        elif other.len() == 1:
-            out = list(self.value < other.value)
-        else:
-            self._compareLengths(other)
-            out = [elemA < elemB for elemA, elemB in zip(self.value, other.value)]
+        out = self.value < other.value
         
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
 
     def __le__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self <= variable(other, self.unit)
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
             raise ValueError(f'You cannot compare {self} and {other} as they do not have the same SI base unit')
@@ -593,22 +676,14 @@ class variable():
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
         
-        if self.len() == 1 and other.len() == 1:
-            out = self.value <= other.value
-        elif self.len() == 1:
-            out = list(self.value <= other.value)
-        elif other.len() == 1:
-            out = list(self.value <= other.value)
-        else:
-            self._compareLengths(other)
-            out = [elemA <= elemB for elemA, elemB in zip(self.value, other.value)]
-
+        out = self.value <= other.value
+        
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
     
     def __gt__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self > variable(other, self.unit)
 
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
@@ -620,22 +695,14 @@ class variable():
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
 
-        if self.len() == 1 and other.len() == 1:
-            out =  self.value > other.value
-        elif self.len() == 1:
-            out =  list(self.value > other.value)
-        elif other.len() == 1:
-            out =  list(self.value > other.value)
-        else:
-            self._compareLengths(other)
-            out =  [elemA > elemB for elemA, elemB in zip(self.value, other.value)]
-        
+        out = self.value > other.value
+           
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
    
     def __ge__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self >= variable(other, self.unit)
 
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
@@ -647,22 +714,14 @@ class variable():
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
 
-        if self.len() == 1 and other.len() == 1:
-            out =  self.value >= other.value
-        elif self.len() == 1:
-            out =  list(self.value >= other.value)
-        elif other.len() == 1:
-            out =  list(self.value >= other.value)
-        else:
-            self._compareLengths(other)
-            out =  [elemA >= elemB for elemA, elemB in zip(self.value, other.value)]
-
+        out = self.value >= other.value
+        
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
 
     def __eq__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self == variable(other, self.unit)
 
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
@@ -673,23 +732,18 @@ class variable():
         
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
-
-        if self.len() == 1 and other.len() == 1:
-            out =  self.value == other.value
-        elif self.len() == 1:
-            out =  list(self.value == other.value)
-        elif other.len() == 1:
-            out =  list(self.value == other.value)
-        else:
-            self._compareLengths(other)
-            out =  [elemA == elemB for elemA, elemB in zip(self.value, other.value)]
-
+        
+        if isinstance(self, arrayVariable) and isinstance(other, arrayVariable):
+            if len(self) != len(other):
+                raise ValueError(f"operands could not be broadcast together with shapes {self.value.shape} {other.value.shape}")    
+        out = self.value == other.value
+            
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
 
     def __ne__(self, other):
-        if not isinstance(other, variable):
+        if not isinstance(other, scalarVariable):
             return self != variable(other, self.unit)
 
         if not self._unitObject._SIBaseUnit == other._unitObject._SIBaseUnit:
@@ -701,16 +755,11 @@ class variable():
         self.convert(self._unitObject._SIBaseUnit)
         other.convert(self._unitObject._SIBaseUnit)
 
-        if self.len() == 1 and other.len() == 1:
-            out = self.value != other.value
-        elif self.len() == 1:
-            out = list(self.value != other.value)
-        elif other.len() == 1:
-            out = list(self.value != other.value)
-        else:
-            self._compareLengths(other)
-            out = [elemA != elemB for elemA, elemB in zip(self.value, other.value)]
-
+        if isinstance(self, arrayVariable) and isinstance(other, arrayVariable):
+            if len(self) != len(other):
+                raise ValueError(f"operands could not be broadcast together with shapes {self.value.shape} {other.value.shape}")    
+        out = self.value != other.value
+            
         self.convert(selfUnit)
         other.convert(otherUnit)
         return out
@@ -718,65 +767,200 @@ class variable():
     def __hash__(self):
         return id(self)
 
+class arrayVariable(scalarVariable):
+   
+    def __len__(self):
+        return len(self._value)
 
-def implements(numpy_function):
-    """Register an __array_function__ implementation for Physical objects."""
-    def decorator(func):
-        HANDLED_FUNCTIONS[numpy_function] = func
-        return func
-    return decorator
-
-
-@implements(np.max)
-def np_max_for_variable(x, *args, **kwargs):
-    if x.len() > 1:
-        index = np.argmax(x.value)
-        val = x.value[index]
-        if not x.uncert is None:
-            unc = x.uncert[index]
+    def __getitem__(self, index):
+        if isinstance(index, int) or isinstance(index, slice):
+            if index >= len(self):
+                raise IndexError('Index out of bounds')
+            if len(self) == 1:
+                if index != 0:
+                    raise IndexError('Index out of bound')
+                return variable(self.value, self._unitObject, self.uncert)
+            return variable(self.value[index], self._unitObject, self.uncert[index])
         else:
-            unc = None
+            val = [self.value[elem]for elem in index]
+            unc = [self.uncert[elem]for elem in index]
+            return variable(val, self._unitObject, unc)
+ 
+    def __setitem__(self, i, elem):
+        if (type(elem) != scalarVariable):
+            raise ValueError(f'You can only set an element with a scalar variable')
+        if (elem.unit != self.unit):
+            raise ValueError(f'You can not set an element of {self} with {elem} as they do not have the same unit')
+        self._value[i] = elem.value
+        self._uncert[i] = elem.uncert
+    
+    def append(self, elem):
+        if not (isinstance(elem, scalarVariable)):
+            raise ValueError(f'You can only set an element with a variable')
+        if (elem.unit != self.unit):
+            raise ValueError(f'You can not set an element of {self} with {elem} as they do not have the same unit')
+        
+        self._value = np.append(self._value, elem.value)
+        self._uncert = np.append(self._uncert, elem.uncert)
+    
+       
+ 
+    def __str__(self, pretty=None) -> str:
+        unitStr = self._unitObject.__str__(pretty=pretty)
+
+        if pretty:
+            pm = r'\pm'
+            space = r'\ '
+            squareBracketLeft = r'\left ['
+            squareBracketRight = r'\right ]'
+
+        else:
+            pm = '+/-'
+            squareBracketLeft = '['
+            squareBracketRight = ']'
+            space = ' '
+
+        if unitStr == '1':
+            unitStr = ''
+        else:
+            unitStr = rf'{squareBracketLeft}{unitStr}{squareBracketRight}'
+
+        # print array of values
+        valStr = []
+        uncStr = []
+        for v, u in zip(self.value, self.uncert):
+            v, u = self.printUncertanty(v, u)
+            valStr.append(v)
+            uncStr.append(u)
+
+        if all(self._uncert == 0) or all(elem is None for elem in self._uncert):
+            out = rf''
+            out += rf'['
+            for i, elem in enumerate(valStr):
+                out += rf'{elem}'
+                if i != len(valStr) - 1:
+                    out += rf', '
+            out += rf']'
+            out += rf'{space}{unitStr}'
+            return out
+        else:
+            # find number of significant digits in uncertanty
+            out = rf''
+            out += rf'['
+            for i, elem in enumerate(valStr):
+                out += rf'{elem}'
+                if i != len(valStr) - 1:
+                    out += r', '
+            out += rf']'
+            out += rf' {pm} '
+            out += rf'['
+            for i, elem in enumerate(uncStr):
+                out += rf'{elem}'
+                if i != len(uncStr) - 1:
+                    out += r', '
+            out += rf']'
+            out += rf'{space}{unitStr}'
+            return out
+    
+    def __pow__(self, other):
+        if not (isinstance(other, scalarVariable) or isinstance(other, arrayVariable)):
+            return self ** variable(other)
+    
+        if isinstance(other, arrayVariable):
+            if len(self) != len(other):
+                raise ValueError(f'operands could not be broadcast together with shapes {self.value.shape} {other.value.shape}')
+            out = [a**b for a,b in zip(self, other)]
+        else:
+            out = [a**other for a in self]
+
+        if all([out[0].unit == elem.unit for elem in out]):
+            return variable([elem.value for elem in out], out[0].unit, [elem.uncert for elem in out])
+        
+        return out
+     
+    def __rpow__(self,other):
+        if not (isinstance(other, scalarVariable) or isinstance(other, arrayVariable)):
+            return variable(other) ** self
+    
+        if isinstance(other, arrayVariable):
+            if len(self) != len(other):
+                raise ValueError(f'operands could not be broadcast together with shapes {self.value.shape} {other.value.shape}')
+            out = [a**b for a,b in zip(other, self)]
+        else:
+            out = [other**a for a in self]
+        
+        if all([out[0].unit == elem.unit for elem in out]):
+            return variable([elem.value for elem in out], out[0].unit, [elem.uncert for elem in out])
+        return out        
+           
+    def __array_ufunc__(self, ufunc, *args, **kwargs):
+        match ufunc:
+            case np.log:
+                return self.log()
+            case np.log10:
+                return self.log10()
+            case np.exp:
+                return self.exp()
+            case np.sin:
+                return self.sin()
+            case np.cos:
+                return self.cos()
+            case np.tan:
+                return self.tan()
+
+    def min(self):
+        index = np.argmin(self.value)
+        return variable(self.value[index], self.unit, self.uncert[index])
+
+    def max(self):
+        index = np.argmax(self.value)
+        return variable(self.value[index], self.unit, self.uncert[index])
+    
+    def mean(self):
+        return sum(self) / len(self)
+
+
+
+
+def variable(value, unit = '', uncert = None, nDigits = 3):
+    # store the value and the uncertaty
+    def evaluateInput(input):
+        if input is None:
+            return input
+        if isinstance(input, np.ndarray):
+            if len(input) == 1:
+                return input[0]
+            return input
+        else:
+            if isinstance(input, list):
+                if len(input) == 1:
+                    return input[0]
+                return np.array(input, dtype=float)
+            else:
+                return float(input)
+                
+    value = evaluateInput(value)
+    uncert = evaluateInput(uncert)
+    
+    if uncert is None:
+        if isinstance(value,np.ndarray):
+            uncert = np.zeros(len(value))
+        else:
+            uncert = 0
     else:
-        val = x.value
-        if not x.uncert is None:
-            unc = x.uncert
+        if isinstance(value,np.ndarray):
+            if not isinstance(uncert, np.ndarray):
+                raise ValueError('The lenght of the value has to be equal to the lenght of the uncertanty')
+            if len(uncert) != len(value):
+                raise ValueError('The lenght of the value has to be equal to the lenght of the uncertanty')
         else:
-            unc = None
-    return variable(val, x.unit, unc)
-
-
-@implements(np.min)
-def np_max_for_variable(x, *args, **kwargs):
-    if x.len() > 1:
-        index = np.argmin(x.value)
-        val = x.value[index]
-        if not x.uncert is None:
-            unc = x.uncert[index]
-        else:
-            unc = None
+            if isinstance(uncert, np.ndarray):
+                raise ValueError('The lenght of the value has to be equal to the lenght of the uncertanty')      
+    
+    if isinstance(value, np.ndarray):
+        return arrayVariable(value, unit, uncert, nDigits)
     else:
-        val = x.value
-        if not x.uncert is None:
-            unc = x.uncert
-        else:
-            unc = None
-    return variable(val, x.unit, unc)
+        return scalarVariable(value, unit, uncert, nDigits)
 
 
-@implements(np.mean)
-def np_mean_for_variable(x, *args, **kwargs):
-    if x.len() > 1:
-        val = np.mean(x.value)
-        if not x.uncert is None:
-            n = len(x.uncert)
-            unc = np.sqrt(sum([(1 / n * elem)**2 for elem in x.uncert]))
-        else:
-            unc = None
-    else:
-        val = x.value
-        if not x.uncert is None:
-            unc = x.uncert
-        else:
-            unc = None
-    return variable(val, x.unit, unc)
-
+## TODO add a method to add custom units
